/*! \file bitimage.c

  \brief bit-image handring functions

  \author
    Copyright (C) SEIKO EPSON CORPORATION 2010.

    Driver for Epson SIDM Printers

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>             /* for memset(3) */

#include "output.h"
#include "bitimage.h"
#include "debug.h"


/*
 * proto type define
 */

LOCALFUNC int band_out( bm_info_t *bm_info, unsigned char *bmpbuf, int mode, int band_bytes, int start, int length, unsigned int xreso, unsigned long hmove_base);



/*! \brief n line feed
  \param  vreso         Virtical resolution
  \param  skip_line_num Number of skips
  \param  vmove_base    Vertical move base unit for "ESC J"
  \return               error status
  \retval 0             completd
  \retval other         error
 */
FUNCTION int line_feed( unsigned int vreso, unsigned short skip_line_num, unsigned long vmove_base )
{
  int                       write_error = 0;
  unsigned short            skip_line = 0;
  unsigned char             buf[1] = {0,};

  if ( ARGERROR( (180 != vmove_base) && (216 != vmove_base) ) ) { return EINVAL; }
  if ( ARGERROR(    0 ==  vreso         ) ) { return EINVAL; }
  if ( ARGERROR(    0 ==  skip_line_num ) ) { return 0; } /* fail safe */

  skip_line = my_round_ul(skip_line_num, vreso, vmove_base);
  while ( skip_line ) {
    write_error |= printf_out("%c%s", ESC, "J");
    if ( 255 > skip_line ) {
      buf[0] = (unsigned char)skip_line;
    } else {
      buf[0] = 255;
    }
    write_error |= printd_out(buf, 1);

    skip_line = skip_line - (unsigned short)buf[0];
  }

  if ( write_error ) {
    DEBUG_ERROR_TRACE;
    return EIO;
  }
  return 0;
}



/*! \brief Get bit image mode

    \todo   multi-pass Pseudo High Resolution print is not support yet.

    \param  xreso               horizontal resolution
    \param  yreso               vertical resolution
    \param  dots_per_column     number of dots per column
    \param  high_speed_mode     high speed mode status flag
    \param  support_pse_bmp    'Print Speed Enhancer' support status flag

    \param  escp_bitimage_mode  data store pointer for ESC/P bit-image mode (for 'ESC * m')
    \param  npass               data store pointer for number of pass (for pseudo high-resolution print)

    \return 0     success
    \retval other error
 */
FUNCTION int get_bitimage_mode( unsigned int xreso, unsigned int yreso, unsigned int dots_per_column, int high_speed_mode, unsigned short support_pse_bmp, unsigned char *escp_bitimage_mode, int *npass )
{
  /*!< \brief resolution table */
  typedef struct resolution_table_s {
    unsigned int    yreso;  /*!< User specified Virtical   resolution */
    unsigned int    xreso;  /*!< User specified Horizontal resolution */
    unsigned int    dpc;    /*!< Dots per column */
    int             highsp; /*!< Enable High-Speed mode */
    unsigned short  pse;    /*!< Enable Print Speed Enhancer (category 1 only) */
    unsigned char   mode;   /*!< bit-image data mode for ESC * m */
    int             pass;   /*!< number of pass for pseudo high-resolution print */
  } resolution_table_t;

  static const resolution_table_t reso_tbl[] = {
    /* yreso,xreso,dpc,highsp,  pse, mode, pass */
    {     60,   60, 24,   0,       0,   0,   1 },
    {     60,  120, 24,   0,       0,   1,   1 },
    {     60,  120, 24,   1,       0,   2,   1 },
    {     60,  240, 24,   0,       0,   3,   1 },
    {     60,   80, 24,   0,       0,   4,   1 },
    {     60,   90, 24,   0,       0,   6,   1 },
    {     72,   60,  8,   0,       0,   0,   1 },
    {     72,  120,  8,   0,       0,   1,   1 },
    {     72,  120,  8,   1,       0,   2,   1 },
    {     72,  240,  8,   0,       0,   3,   1 },
    {     72,   80,  8,   0,       0,   4,   1 },
    {     72,   72,  8,   0,       0,   5,   1 },
    {     72,   90,  8,   0,       0,   6,   1 },
    {     72,  144,  8,   0,       0,   7,   1 },
#if 0
    {    144,  120,  8,   0,       0,   1,   2 }, /* multi-pass Pseudo High Resolution print */
    {    144,  240,  8,   0,       0,   3,   2 }, /* multi-pass Pseudo High Resolution print */
#endif
    {    180,   60, 24,   0,       0,  32,   1 },
    {    180,  120, 24,   1,  0x0001,  34,   1 }, /* Print Speed Enhancer (BMP) */
    {    180,  120, 24,   0,       0,  33,   1 },
    {    180,   90, 24,   0,       0,  38,   1 },
    {    180,  180, 24,   1,  0x0001,  37,   1 }, /* Print Speed Enhancer (BMP) */
    {    180,  180, 24,   0,       0,  39,   1 },
    {    180,  360, 24,   0,       0,  40,   1 },
    {    180,  240, 24,   1,  0x0001,  35,   1 }, /* Print Speed Enhancer (BMP) */
    {    360,   60, 48,   0,       0,  64,   1 },
    {    360,  120, 48,   0,       0,  65,   1 },
    {    360,   90, 48,   0,       0,  70,   1 },
    {    360,  180, 48,   0,       0,  71,   1 },
    {    360,  360, 48,   0,       0,  73,   1 }
  };
  static const int num_reso_data = sizeof(reso_tbl) / sizeof(resolution_table_t);

  int i;

  if ( ARGERROR( NULL == escp_bitimage_mode ) ) { return EINVAL; }
  if ( ARGERROR( NULL == npass              ) ) { return EINVAL; }

  /* 1st try */
  for ( i = 0; i < num_reso_data; i++ ) {

      if ( ( reso_tbl[i].highsp == high_speed_mode ) &&
           ( reso_tbl[i].dpc    == dots_per_column ) &&
           ( reso_tbl[i].xreso  == xreso ) &&
           ( reso_tbl[i].yreso  == yreso ) ) {

        /* pse check */
        if ( 0 != reso_tbl[i].pse ) {
          /* pse only mode */
          if ( 0 == (reso_tbl[i].pse & support_pse_bmp) ) {
            continue;           /* NG : pse only mode */
          }
        }

        *escp_bitimage_mode = reso_tbl[i].mode;
        *npass = reso_tbl[i].pass;
        return 0;               /* found */
      }
    }

  /* 2nd try (without high-speed mode) */
  for ( i = 0; i < num_reso_data; i++ ) {

      if ( ( reso_tbl[i].dpc    == dots_per_column ) &&
           ( reso_tbl[i].xreso  == xreso ) &&
           ( reso_tbl[i].yreso  == yreso ) ) {

        /* pse check */
        if ( 0 != reso_tbl[i].pse ) {
          /* pse only mode */
          if ( 0 == (reso_tbl[i].pse & support_pse_bmp) ) {
            continue;           /* NG : pse only mode */
          }
        }

        *escp_bitimage_mode = reso_tbl[i].mode;
        *npass = reso_tbl[i].pass;
        return 0;               /* found */
      }
    }

  /* Give-up */
  DEBUG_ERROR_TRACE;
  return 1;
}


/*! \brief Check resolution

    \param  xreso               horizontal resolution
    \param  yreso               vertical resolution
    \param  dots_per_column     number of dots per column
    \param  high_speed_mode     high speed mode status flag
    \param  support_pse_bmp    'Print Speed Enhancer' support status flag
    \retval 1    success (supported)
    \retval 0    error (not supported)
 */
FUNCTION int is_supported_resolution( unsigned int xreso, unsigned int yreso, unsigned int dots_per_column, int high_speed_mode, unsigned short support_pse_bmp)
{
  unsigned char escp_bitimage_mode = 0;
  int npass = 0;

  if ( 0 == get_bitimage_mode( xreso, yreso, dots_per_column, high_speed_mode,
                               support_pse_bmp, &escp_bitimage_mode, &npass ) ) {
    return 1;                   /* success (supported) */
  }

  DEBUG_ERROR_TRACE;
  return 0;                     /* error (not supported) */
}



/*! \brief Band data send to printer
  \param  bm_info       bit-image data information #bm_info_t
  \param  band_line     Number of line of raster data in the \a band_buf
  \param  band_buf      Band buffer for print image
  \return               error status
  \retval 0             completd
  \retval other         error
 */
FUNCTION int send_band( bm_info_t *bm_info, unsigned short band_line, const unsigned char *band_buf )
{
  unsigned char            *bmbuf = NULL;
  unsigned                  raster_band_bytes_per_line = 0;
  int                       bitmap_band_hight_bytes = 0;
  long                      bitmap_buf_size = 0;
  int                       ref_pos = 0;
  int                       set_pos = 0;
  unsigned char             maskbit = 0;
  unsigned char             set_bit = 0;
  int                       x = 0;
  int                       y = 0;
  int                       write_error = 0;
  int                       status = 0;
  int                       black_start_pixel = -1;
  int                       black_end_pixel = -1;
  unsigned char             escp_bitimage_mode = 0;
  int                       npass = 0;

  if ( ARGERROR( NULL ==  bm_info   ) ) { return EINVAL; }
  if ( ARGERROR(    0 ==  band_line ) ) { return EINVAL; }
  if ( ARGERROR( NULL ==  band_buf  ) ) { return EINVAL; }

  /* bm_info check */
  if ( ARGERROR(    0 == bm_info->band_pixel_per_line ) ) { return 0; } /* fale safe */
  if ( ARGERROR(    0 == bm_info->hmove_base          ) ) { return EINVAL; }
  if ( ARGERROR( NULL == bm_info->bitmap_band_buf     ) ) { return EINVAL; }
  if ( ARGERROR( NULL == bm_info->bmp_data_size       ) ) { return EINVAL; }
  if ( ARGERROR(    0 == bm_info->dots_per_column     ) ) { return EINVAL; }
  if ( ARGERROR(    0 == bm_info->HWResolution_H      ) ) { return EINVAL; }
  if ( ARGERROR(    0 == bm_info->HWResolution_V      ) ) { return EINVAL; }

  ASSERT(0 == (bm_info->dots_per_column % 8));

  /* bit-image data max size check (ESC/P command spec regulation) */
  if ( ARGERROR((32 * 256) <= bm_info->band_pixel_per_line) ) {
    /* bit-image data too big! */
    return -1;
  }

  status = get_bitimage_mode(bm_info->HWResolution_H,
                             bm_info->HWResolution_V,
                             bm_info->dots_per_column,
                             bm_info->high_speed_mode,
                             bm_info->support_pse_bmp,
                             &escp_bitimage_mode, &npass );
  if ( 0 != status ) {
    DEBUG_ERROR_TRACE;
    fprintf(stderr, "ERROR: Illegal Resolution.\n");
    if (1) { debug_log(dbg, "ERROR: The resolution (%ux%u) is not supported.\n",
                       bm_info->HWResolution_H,
                       bm_info->HWResolution_V); }
    return -1;
  }

  if (1) { debug_log(dbg, "bitimage output: width=%u hight=%u mode=%d(0x%02x) npass=%d\n",
                     bm_info->band_pixel_per_line,
                     bm_info->dots_per_column,
                     escp_bitimage_mode,
                     escp_bitimage_mode, npass); }


  /* :TODO: Multi-pass print is not support yet. */
  {

    raster_band_bytes_per_line = (bm_info->band_pixel_per_line + 7) / 8;
    bitmap_band_hight_bytes = (bm_info->dots_per_column + 7) / 8;

    bitmap_buf_size = bitmap_band_hight_bytes * bm_info->band_pixel_per_line;

    bmbuf = my_malloc(bitmap_buf_size);
    if ( NULL == bmbuf ) {
      DEBUG_ERROR_TRACE;
      perror("ERROR");
      return ENOMEM;
    }
    *(bm_info->bitmap_band_buf) = bmbuf;

    memset(bmbuf, 0, bitmap_buf_size);

    black_start_pixel = -1;
    black_end_pixel = -1;

    maskbit = 0;
    set_bit = 0;

    for ( x = 0; x < bm_info->band_pixel_per_line; x++ ) {

      if ( !maskbit ) {
        maskbit = 0x80;
      }

      for ( y = 0; y < bm_info->dots_per_column; y++ ) {
        ref_pos = (y * raster_band_bytes_per_line) + (x / 8);
        set_pos = (x * bitmap_band_hight_bytes) + (y / 8);

        if ( !set_bit ) {
          set_bit = 0x80;
        }

        if ( y < band_line ) {
          if ( band_buf[ref_pos] & maskbit ) {
            bmbuf[set_pos] |= set_bit;
            if ( 0 > black_start_pixel ) {
              black_start_pixel = x;
            }
            black_end_pixel = x;
          }
        }
        set_bit = set_bit >> 1;
      }
      maskbit = maskbit >> 1;
    }

    ASSERT(0 <= black_end_pixel);
    ASSERT(0 <= black_start_pixel);

    if ( ( 0 <= black_start_pixel) && ( 0 <= black_end_pixel ) ) {
      status = band_out(bm_info, bmbuf,
                        escp_bitimage_mode,
                        bitmap_band_hight_bytes,
                        black_start_pixel,
                        black_end_pixel - black_start_pixel + 1,
                        bm_info->HWResolution_H,
                        bm_info->hmove_base);
    }

    my_free(bmbuf);
    bmbuf = NULL;
    *(bm_info->bitmap_band_buf) = NULL;

#if 1
    /* current position move to lefthand. */
    write_error |= printf_out("%c", CR);
#endif

  }

  if ( 0 != status ) {
    DEBUG_ERROR_TRACE;
    return status;
  }

  if ( write_error ) {
    DEBUG_ERROR_TRACE;
    return EIO;
  }
  return 0;
}


/*! \brief Output one band bitmap image data to printer.
  \param  bm_info       bit-image data information #bm_info_t
  \param  bmpbuf        bitmap image data buffer
  \param  mode          bit-image mode (resolution)
  \param  band_bytes    bitmap image data height (in Bytes)
  \param  start         black bitmap image start x point (in pixel)
  \param  length        black bitmap image length (in pixel)
  \param  xreso         horizontal resolution the bitmap image data
  \param  hmove_base    horizontal move base unit for "ESC $"
  \return               error status
  \retval 0             completed
  \retval other         error
 */
LOCALFUNC int band_out( bm_info_t *bm_info, unsigned char *bmpbuf, int mode, int band_bytes, int start, int length, unsigned int xreso, unsigned long hmove_base)
{
  unsigned char     buf[4] = {0,};
  int               write_error = 0;
  unsigned char    *sendbuf = NULL;
  long              lnum = 0;

  if ( ARGERROR(     NULL == bm_info    ) ) { return EINVAL; }
  if ( ARGERROR(     NULL == bmpbuf     ) ) { return EINVAL; }
  if ( ARGERROR(        0 >  mode       ) ) { return EINVAL; }
  if ( ARGERROR(        0 >= band_bytes ) ) { return EINVAL; }
  if ( ARGERROR(        0 >  start      ) ) { return EINVAL; }
  if ( ARGERROR(        0 >= length     ) ) { return EINVAL; }
  if ( ARGERROR( 32 * 256 <= length     ) ) { return EINVAL; } /* ESC/P regulation */
  if ( ARGERROR(        0 == xreso      ) ) { return EINVAL; }
  if ( ARGERROR(        0 == hmove_base ) ) { return EINVAL; }

  sendbuf = &bmpbuf[band_bytes * start];
  lnum = my_round_ul(start, xreso, hmove_base);

  write_error |= printf_out("%c%s", ESC, "$");
  buf[0] = lnum % 256;
  buf[1] = lnum / 256;
  write_error |= printd_out(buf, 2);

  *(bm_info->bmp_data_size) = band_bytes * length; /* for job cancel */

  write_error |= printf_out("%c%s", ESC, "*");
  buf[0] = mode;
  buf[1] = length % 256;
  buf[2] = length / 256;
  write_error |= printd_out(buf, 3);

  write_error |= printd_out(sendbuf, band_bytes * length);

  *(bm_info->bmp_data_size) = 0; /* for job cancel */

  if ( write_error ) {
    DEBUG_ERROR_TRACE;
    return EIO;
  }

  return 0;
}

/* end of file */
