/*! \file option.c

  \brief Option keyword and value information

  \author
    Copyright (C) SEIKO EPSON CORPORATION 2010.

    Driver for Epson SIDM Printers

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>             /* for strcmp(3) */

#include "option.h"
#include "debug.h"

/*
 * File scope global parameters...
 */

/*! \brief option key list */
static key_value_t option_list[] = {
  { PPDKEYPageSize      , "" },
  { PPDKEYResolution    , "" },
  { PPDKEYPrintDirection, "" },
  { PPDKEYTopMP         , "" },
  { PPDKEYPacketMode    , "" },
  { PPDKEYSpeed         , "" },
  { PPDKEYInputSlot     , "" },
  { PPDKEYSvRP          , "" },
  { PPDKEYTearOff       , "" },
  { PPDATTRSpPseBmp     , "" },
  { PPDATTRSpPseFsx     , "" }
};

/*! \brief number of key/value in the option_list */
static const int option_list_num = sizeof(option_list) / sizeof(key_value_t);



/*! \brief Find option choice
  \param key
  \return            pointer for the option.
  \retval NULL       the \a key is not found.
 */
LOCALFUNC key_value_t *find_option( const char *key )
{
  key_value_t *options;
  int i;

  if ( ARGERROR( NULL == key ) ) { return NULL; }

  options = option_list;

  for ( i = 0; i < option_list_num; i++) {
    if ( 0 == strcmp(key, options->key) ) {
      return options;
    }
    options++;
  }
  return NULL;
}

/*! \brief Find option choice
  \param key
  \return            value for the \a key.
  \retval NULL       the \a key is not found, or the \a key have no value.
 */
FUNCTION char *find_key_value( const char *key )
{
  key_value_t *found_option;

  if ( ARGERROR( NULL == key ) ) { return NULL; }

  found_option = find_option(key);
  if ( NULL == found_option ) {
    return NULL;
  }

  if ( 0 == found_option->value[0] ) {
    return NULL;
  }

  return found_option->value;
}

/*! \brief Check value in the option list.
  \param key
  \param value
  \return         match status.
  \retval 1       match
  \retval 0       no match
 */
FUNCTION int is_match_key_value( const char *key, char *value )
{
  const char *found_value = NULL;

  if ( ARGERROR( NULL == key         ) ) { return 0; }
  if ( ARGERROR( NULL == value       ) ) { return 0; }

  found_value = find_key_value(key);
  if ( NULL != found_value ) {
    if ( 0 == strcmp(value, found_value) ) {
      return 1;                 /* matched */
    }
  }
  return 0;                     /* un-matched */
}


/*! \brief Check head of value in the option list.
  \param key      PPD key string
  \param value    Value of \a key
  \param num      Number of bytes of head of a string
  \return         match status.
  \retval 1       match
  \retval 0       no match
 */
FUNCTION int is_match_key_value_n( const char *key, char *value, int num )
{
  const char *found_value = NULL;

  if ( ARGERROR( NULL == key         ) ) { return 0; }
  if ( ARGERROR( NULL == value       ) ) { return 0; }
  if ( ARGERROR(    0 >= num         ) ) { return 0; }

  found_value = find_key_value(key);
  if ( NULL != found_value ) {
    if ( 0 == strncmp(found_value, value, num) ) {
      return 1;                 /* matched */
    }
  }
  return 0;                     /* un-matched */
}


/*! \brief Set current PPD file options to the option list.
  \param ppd              CUPS ppd_file_t
  \retval 0          success
  \retval other      error
 */
FUNCTION int set_options( ppd_file_t *ppd )
{
  key_value_t  *key_value;
  ppd_option_t *ppd_option;
  ppd_choice_t *marked_choice;

  FUNCSTART;

  if ( ARGERROR( NULL == ppd         ) ) { return EINVAL; }

  for ( ppd_option = ppdFirstOption(ppd); ppd_option != NULL; ppd_option = ppdNextOption(ppd)) {

    key_value = find_option(ppd_option->keyword);
    if ( NULL != key_value ) {

      marked_choice = ppdFindMarkedChoice(ppd, key_value->key);

      if ( NULL == marked_choice ) {
        strncpy(key_value->value, ppd_option->defchoice, PPD_MAX_NAME -1);
      } else {
        strncpy(key_value->value, marked_choice->choice, PPD_MAX_NAME -1);
      }
    }
  }

  return 0;
}


#if (0 != MYDEBUG)
/*! \brief show key and value list

    Key and value list output to debug log file.

    \see #debug_open()
 */
FUNCTION void debug_show_option_list( void )
{
  int i;

  dbgprintf("\n* Key value list START *************************************\n");

  for ( i = 0; 0 != option_list[i].key[0]; i++ ) {
    dbgprintf("%s=%s\n", option_list[i].key, option_list[i].value);
  }

  dbgprintf("num_options = %d\n", i);
  dbgprintf("* Key value list END ***************************************\n");
}
#endif  /* (0 != MYDEBUG) */

/* end of file */
