/*! \file output.c

  \brief output utility.

  \author
    Copyright (C) SEIKO EPSON CORPORATION 2010.

    Driver for Epson SIDM Printers

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdarg.h>
#include "output.h"
#include "debug.h"

/*! \brief (for DEBUG only)

  Output delay count in micro sec.
  \note Set 50000 (50ms delay), When job cancel test.
 */
#define  OUTPUT_DELAY 0



/* ****************************************************************
   standard C function wrapper
   **************************************************************** */

/*! \brief formated text message output to printer.

    Like a printf(3).

    \param  fmt    Formated message string, like a printf(3).
    \param  ...    Argument params for \a fmt, like a printf(3).
    \return        Output status.
    \retval 0      completed
    \retval other  error
 */
FUNCTION int printf_out( const char *fmt, ... )
{
  va_list ap;
  int n = 0;

  if ( ARGERROR( NULL == fmt ) ) { return EINVAL; }

  va_start(ap, fmt);
  n = vfprintf(stdout, fmt, ap);
  va_end(ap);

  if ( 0 > n ) {
    return 1;
  }

  return 0;
}



/*! \brief binary data output.

    Binary data output to printer.
    Like a \c fwrite(3).

    \param  data  binary data.
    \param  size  output binary data size in Bytes.
    \return 0     completed
    \retval other write error.

 */
FUNCTION int printd_out( const unsigned char *data, size_t size )
{
  size_t writed = 0;
  size_t n = 0;
  int i = 0;

  if ( ARGERROR( NULL == data ) ) { return EINVAL; }
  if ( ARGERROR(    0 == size ) ) { return EINVAL; }

  if ( 0 != MYDEBUG ) {
    for ( i = 0; i < size; i++ ) {
      fputc(data[i], stdout);

      if ( OUTPUT_DELAY ) {     /* (for job cancel test = 50ms) */
        safe_usleep(OUTPUT_DELAY);
      }
    }
    n = i;

  } else {

    n = 0;
    for ( i = 0; (3 > i) && (n < size); i++ ) { /* retry limit = 3 */
      writed = fwrite(&data[n], sizeof(unsigned char), (size - n), stdout);
      n += writed;
    }
  }

  if ( n != size ) {
    return 1;
  }
  return 0;
}

/* end of file */
