/*! \file page.c

  \brief page handring functions

  \author
    Copyright (C) SEIKO EPSON CORPORATION 2010.

    Driver for Epson SIDM Printers

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <signal.h>

#include "escpcmd.h"
#include "jobinfo.h"
#include "page.h"
#include "bitimage.h" /* for is_supported_resolution() */
#include "debug.h"


/*! \brief Start Page
  \param  job_info      job information
  \param  page_header   page information
  \return               error status
  \retval 0             completd
  \retval other         error
 */
FUNCTION int start_page( job_info_t *job_info, const cups_page_header2_t *page_header )
{
  escp_page_info_t      escp_page_info;
  int                   write_error = 0;
  int                   high_speed = -1; /* 1:FSx1, 0:FSx0, -1:no send FSx */
  const char           *opt_speed = NULL;
  unsigned char         escp_fsx_value = 0;

#if defined(HAVE_SIGACTION) && !defined(HAVE_SIGSET)
  struct sigaction action;                      /* Actions for POSIX signals */
#endif /* HAVE_SIGACTION && !HAVE_SIGSET */

  FUNCSTART;

  if ( ARGERROR( NULL == job_info    ) ) { return EINVAL; }
  if ( ARGERROR( NULL == page_header ) ) { return EINVAL; }

  /*
   * Register a signal handler to eject the current page if the
   * job is cancelled.
   */

#ifdef HAVE_SIGSET /* Use System V signals over POSIX to avoid bugs */
  sigset(SIGTERM, cancel_job);
#elif defined(HAVE_SIGACTION)
  memset(&action, 0, sizeof(action));

  sigemptyset(&action.sa_mask);
  action.sa_handler = cancel_job;
  sigaction(SIGTERM, &action, NULL);
#else
  signal(SIGTERM, cancel_job);
#endif /* HAVE_SIGSET */


  if ( 1 != job_info->page_num ) {

    write_error = escp_start_page();

  } else {                      /* first page init */

    job_info->band_pixel_per_line = (job_info->page_size.right - job_info->page_size.left) * page_header->HWResolution[0] / 72;
    job_info->band_bytes_per_line = (job_info->band_pixel_per_line + 7) / 8;

    /* Failsafe for input of image data of illegal size from CUPS.  */
    if ( page_header->cupsBytesPerLine > job_info->band_bytes_per_line ) {
      dbgprintf("WARNING: cupsBytesPerLine > PrintableAreaWidth\n");
      dbgprintf("         Input CUPS Raster image data width = %u Bytes\n", page_header->cupsBytesPerLine);
      dbgprintf(" correct output ESC/P  bit image data width = %u Bytes\n", job_info->band_bytes_per_line);
    }

    /* Failsafe for input of image data of illegal size from CUPS.  */
    if ( job_info->band_bytes_per_line > page_header->cupsBytesPerLine ) {
      job_info->band_bytes_per_line = page_header->cupsBytesPerLine;
      job_info->band_pixel_per_line = job_info->band_bytes_per_line * 8;
      dbgprintf("WARNING: cupsBytesPerLine < PrintableAreaWidth\n");
      dbgprintf("         Input CUPS Raster image data width = %u Bytes\n", page_header->cupsBytesPerLine);
      dbgprintf(" correct output ESC/P  bit image data width = %u Bytes\n", job_info->band_bytes_per_line);
    }


    job_info->lines_per_page = (job_info->page_size.top - job_info->page_size.bottom) * page_header->HWResolution[1] / 72;

    /* Failsafe for input of image data of illegal size from CUPS.  */
    if ( job_info->lines_per_page < page_header->cupsHeight ) {
      dbgprintf("WARNING: cupsHight > PrintableAreaHight\n");
      dbgprintf("         Input CUPS Raster image data hight = %u Lines\n", page_header->cupsHeight);
      dbgprintf(" correct output ESC/P  bit image data hight = %u Lines\n", job_info->lines_per_page);
    }

    /* Failsafe for input of image data of illegal size from CUPS.  */
    if ( job_info->lines_per_page > page_header->cupsHeight ) {
      dbgprintf("WARNING: cupsHight < PrintableAreaHight\n");
      dbgprintf("         Input CUPS Raster image data hight = %u Lines\n", page_header->cupsHeight);
      dbgprintf("         output ESC/P  bit image data hight = %u Lines\n", job_info->lines_per_page);
    }


    /* debug: Show CUPS pageSize parameters. */
    if (1) { debug_show_cups_page_size( &job_info->page_size ); }

    escp_page_info.page_length   = (float)  job_info->page_size.length                               * (job_info->design_base_reso / 72);
    escp_page_info.margin_top    = (float) (job_info->page_size.length - job_info->page_size.top   ) * (job_info->design_base_reso / 72);
    escp_page_info.margin_bottom = (float) (job_info->page_size.length - job_info->page_size.bottom) * (job_info->design_base_reso / 72);
    escp_page_info.margin_left   = (float)  job_info->page_size.left                                 * (job_info->design_base_reso / 72);
    escp_page_info.margin_right  = (float)  job_info->page_size.width                                * (job_info->design_base_reso / 72);

    if ( ARGERROR( (0 > escp_page_info.margin_top) || (escp_page_info.margin_top > escp_page_info.margin_bottom) ) ) {
      escp_page_info.margin_top = 0; /* top margin force set zero */
    }

    /* debug: Show ESC/P page information parameters. */
    if (1) { debug_show_escp_page_size( &escp_page_info, job_info->design_base_reso); }


    /* default high speed mode */
    job_info->high_speed_mode = 0;

    opt_speed   = find_key_value(PPDKEYSpeed);
    if ( NULL != opt_speed ) {

      if ( 0 == strcmp(opt_speed, "Default") ) {
        high_speed = -1;

      } else {

        /* High-Speed mode */
        if ( NULL != strstr(opt_speed, "High") ) {

          high_speed = 1;
          job_info->high_speed_mode = 1;
          escp_fsx_value = 1;

        } else {

          high_speed = 0;
          job_info->high_speed_mode = 0;
          escp_fsx_value = 0;

          ASSERT( (NULL != strstr(opt_speed, "Std")) );
        }
      }
    }

    if ( ! is_supported_resolution(page_header->HWResolution[0],
                                   page_header->HWResolution[1],
                                   page_header->cupsRowCount,
                                   job_info->high_speed_mode,
                                   job_info->support_pse_bmp) ) {
      DEBUG_ERROR_TRACE;
      return EINVAL;
    }

    job_info->input_raster_buff = my_malloc(page_header->cupsBytesPerLine + (job_info->band_bytes_per_line * page_header->cupsRowCount));
    if ( NULL == job_info->input_raster_buff ) {
      DEBUG_ERROR_TRACE;
      perror("ERROR");
      return ENOMEM;
    }
    job_info->band_buf = &(job_info->input_raster_buff[page_header->cupsBytesPerLine]);


    /*
     * Job initialize ESC/P command start...
     */
    write_error = escp_start_job(&escp_page_info, job_info->design_base_reso, job_info->vreso_base, high_speed, job_info->support_pse_fsx);
  }

  /* show page information */
  if (1) { debug_show_raster_info( job_info->page_num,
                                   page_header->cupsWidth,
                                   page_header->cupsHeight,
                                   page_header->cupsRowCount ); }

  /* check output error */
  if ( write_error ) {
    DEBUG_ERROR_TRACE;
    return EIO;
  }

  return 0;
}


/*
 * End of a page printing...
 */

/*! \brief End Page
  \param  job_info      job information
  \return               error status
  \retval 0             completed
  \retval other         error
 */
FUNCTION int end_page( job_info_t *job_info )
{
#if defined(HAVE_SIGACTION) && !defined(HAVE_SIGSET)
  struct sigaction action;                      /* Actions for POSIX signals */
#endif /* HAVE_SIGACTION && !HAVE_SIGSET */

  int                 write_error = 0;

  FUNCSTART;

  if ( ARGERROR( NULL == job_info ) )    { return EINVAL; }

  write_error = escp_end_page();

  if ( NULL != job_info->bitmap_band_buf ) {
    my_free(job_info->bitmap_band_buf);
    job_info->bitmap_band_buf = NULL;
  }

  /*
   * Unregister the signal handler...
   */

#ifdef HAVE_SIGSET /* Use System V signals over POSIX to avoid bugs */
  sigset(SIGTERM, SIG_IGN);
#elif defined(HAVE_SIGACTION)
  memset(&action, 0, sizeof(action));

  sigemptyset(&action.sa_mask);
  action.sa_handler = SIG_IGN;
  sigaction(SIGTERM, &action, NULL);
#else
  signal(SIGTERM, SIG_IGN);
#endif /* HAVE_SIGSET */

  if ( write_error ) {
    DEBUG_ERROR_TRACE;
    return EIO;
  }

  return 0;
}


/* end of file */
